/*
 * Java Genetic Algorithm Library (jenetics-2.0.2).
 * Copyright (c) 2007-2014 Franz Wilhelmstötter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Author:
 *    Franz Wilhelmstötter (franz.wilhelmstoetter@gmx.at)
 */
package org.jenetics;

import java.io.Serializable;
import java.util.Comparator;

import org.testng.Assert;
import org.testng.annotations.Test;

import org.jenetics.util.Function;

/**
 * @author <a href="mailto:franz.wilhelmstoetter@gmx.at">Franz Wilhelmstötter</a>
 * @version <em>$Date: 2014-02-17 $</em>
 */
public class OptimizeTest {

	private static class FF
		implements Function<Genotype<DoubleGene>, Double>,
					Serializable
	{
		private static final long serialVersionUID = 6215331666272441749L;

		@Override
		public Double apply(Genotype<DoubleGene> genotype) {
			return genotype.getGene().getAllele();
		}
	}

	private static final FF _ff = new FF();

	private static Phenotype<DoubleGene, Double> pt(double value) {
		return Phenotype.of(Genotype.of(DoubleChromosome.of(DoubleGene.of(value, 0, 10))), _ff, 0);
	}

	@Test
	public void comparator() {
		Comparator<Phenotype<DoubleGene, Double>> comp =
			Optimize.MAXIMUM.<Phenotype<DoubleGene, Double>>descending();
		Assert.assertTrue(comp.compare(pt(2), pt(3)) > 0);
		Assert.assertTrue(comp.compare(pt(2), pt(2)) == 0);
		Assert.assertTrue(comp.compare(pt(5), pt(3)) < 0);

		comp = Optimize.MINIMUM.<Phenotype<DoubleGene, Double>>descending();
		Assert.assertTrue(comp.compare(pt(4), pt(3)) > 0);
		Assert.assertTrue(comp.compare(pt(2), pt(2)) == 0);
		Assert.assertTrue(comp.compare(pt(2), pt(3)) < 0);
	}

	@Test
	public void compare() {
		final FF ff = new FF();
		final Phenotype<DoubleGene, Double> pt1 = Phenotype.of(Genotype.of(
			DoubleChromosome.of(DoubleGene.of(5, 0, 10))), ff, 0);
		final Phenotype<DoubleGene, Double> pt2 = Phenotype.of(Genotype.of(
			DoubleChromosome.of(DoubleGene.of(7, 0, 10))), ff, 0);
		final Phenotype<DoubleGene, Double> pt3 = Phenotype.of(Genotype.of(
			DoubleChromosome.of(DoubleGene.of(7, 0, 10))), ff, 0);

		Assert.assertTrue(Optimize.MINIMUM.compare(pt1, pt2) > 0);
		Assert.assertTrue(Optimize.MAXIMUM.compare(pt1, pt2) < 0);
		Assert.assertTrue(Optimize.MINIMUM.compare(pt3, pt2) == 0);
		Assert.assertTrue(Optimize.MAXIMUM.compare(pt3, pt2) == 0);
	}

	@Test
	public void best() {
		final FF ff = new FF();
		final Phenotype<DoubleGene, Double> pt1 = Phenotype.of(Genotype.of(
			DoubleChromosome.of(DoubleGene.of(5, 0, 10))), ff, 0);
		final Phenotype<DoubleGene, Double> pt2 = Phenotype.of(Genotype.of(
			DoubleChromosome.of(DoubleGene.of(7, 0, 10))), ff, 0);
		final Phenotype<DoubleGene, Double> pt3 = Phenotype.of(Genotype.of(
			DoubleChromosome.of(DoubleGene.of(7, 0, 10))), ff, 0);

		Assert.assertSame(Optimize.MINIMUM.best(pt1, pt2), pt1);
		Assert.assertSame(Optimize.MAXIMUM.best(pt1, pt2), pt2);
		Assert.assertSame(Optimize.MINIMUM.best(pt2, pt3), pt2);
	}

	@Test
	public void worst() {
		final FF ff = new FF();
		final Phenotype<DoubleGene, Double> pt1 = Phenotype.of(Genotype.of(
			DoubleChromosome.of(DoubleGene.of(5, 0, 10))), ff, 0);
		final Phenotype<DoubleGene, Double> pt2 = Phenotype.of(Genotype.of(
			DoubleChromosome.of(DoubleGene.of(7, 0, 10))), ff, 0);
		final Phenotype<DoubleGene, Double> pt3 = Phenotype.of(Genotype.of(
			DoubleChromosome.of(DoubleGene.of(7, 0, 10))), ff, 0);

		Assert.assertSame(Optimize.MINIMUM.worst(pt1, pt2), pt2);
		Assert.assertSame(Optimize.MAXIMUM.worst(pt1, pt2), pt1);
		Assert.assertSame(Optimize.MINIMUM.worst(pt2, pt3), pt2);
	}

}
